@@ -1,15 +1,15 @@
-import pytest
-import logging
 import keyword
+import logging
 import dataclasses
+from collections import defaultdict
 
+import pytest
 import zigpy.types as zigpy_t
 import zigpy.zdo.types
-import zigpy_znp.commands as c
-import zigpy_znp.frames as frames
-from zigpy_znp import types as t
 
-from collections import defaultdict
+import zigpy_znp.frames as frames
+import zigpy_znp.commands as c
+from zigpy_znp import types as t
 
 
 def test_command_header():
@@ -242,7 +242,11 @@ def test_command_param_binding():
 def test_command_optional_params():
     # Optional params values don't need a value
     short_version_rsp = c.SYS.Version.Rsp(
-        TransportRev=0, ProductId=1, MajorRel=2, MinorRel=3, MaintRel=4,
+        TransportRev=0,
+        ProductId=1,
+        MajorRel=2,
+        MinorRel=3,
+        MaintRel=4,
     )
 
     # Some can still be passed
@@ -332,11 +336,7 @@ def test_command_optional_params_failures():
 
 
 def test_simple_descriptor():
-    # Support both old and new zigpy types
-    try:
-        lvlist16_type = zigpy_t.LVList(t.uint16_t)
-    except TypeError:
-        lvlist16_type = zigpy_t.LVList[t.uint16_t]
+    lvlist16_type = zigpy_t.LVList[t.uint16_t]
 
     simple_descriptor = zigpy.zdo.types.SimpleDescriptor()
     simple_descriptor.endpoint = zigpy_t.uint8_t(1)
@@ -531,3 +531,22 @@ def test_command_replace_partial():
 
     assert command1.replace() == command1
     assert command1.replace(SysId=0x13) == command2
+
+
+def test_command_possibly_empty_payload():
+    class TestSubsystem(t.CommandsBase, subsystem=t.Subsystem.SYS):
+        Test = t.CommandDef(
+            t.CommandType.AREQ,
+            0x00,
+            rsp_schema=(t.Param("Data", t.Bytes, "Can be any length"),),
+        )
+
+    Test = TestSubsystem.Test.Callback
+
+    assert Test.from_frame(
+        frames.GeneralFrame(header=Test.header, data=b"test")
+    ) == Test(Data=t.Bytes(b"test"))
+
+    assert Test.from_frame(frames.GeneralFrame(header=Test.header, data=b"")) == Test(
+        Data=t.Bytes(b"")
+    )
