@@ -6,37 +6,30 @@ import logging
 import warnings
 import itertools
 import contextlib
-import async_timeout
 
 import zigpy.zdo
 import zigpy.util
 import zigpy.types
-import zigpy.device
 import zigpy.config
-import zigpy.profiles
+import zigpy.device
+import async_timeout
 import zigpy.endpoint
+import zigpy.profiles
 import zigpy.zcl.foundation
-
 from zigpy.zcl import clusters
-from zigpy.types import (
-    ExtendedPanId,
-    deserialize as list_deserialize,
-    Struct as ZigpyStruct,
-)
-from zigpy.zdo.types import MultiAddress, ZDOCmd, ZDOHeader, CLUSTERS as ZDO_CLUSTERS
+from zigpy.types import ExtendedPanId, deserialize as list_deserialize
+from zigpy.zdo.types import CLUSTERS as ZDO_CLUSTERS, ZDOCmd, ZDOHeader, MultiAddress
 from zigpy.exceptions import DeliveryError
 
 import zigpy_znp.types as t
 import zigpy_znp.config as conf
 import zigpy_znp.commands as c
-
 from zigpy_znp.api import ZNP
-from zigpy_znp.znp.nib import parse_nib, OldNIB
+from zigpy_znp.znp.nib import NIB, CC2531NIB, parse_nib
 from zigpy_znp.exceptions import InvalidCommandResponse
 from zigpy_znp.types.nvids import NwkNvIds
 from zigpy_znp.zigbee.zdo_converters import ZDO_CONVERTERS
 
-
 with warnings.catch_warnings():
     warnings.filterwarnings(
         action="ignore",
@@ -48,8 +41,20 @@ with warnings.catch_warnings():
 
 
 ZDO_ENDPOINT = 0
+PROBE_TIMEOUT = 5  # seconds
+STARTUP_TIMEOUT = 5  # seconds
 ZDO_REQUEST_TIMEOUT = 10  # seconds
 DATA_CONFIRM_TIMEOUT = 5  # seconds
+NETWORK_COMMISSIONING_TIMEOUT = 30  # seconds
+
+DEFAULT_TC_LINK_KEY = t.TCLinkKey(
+    ExtAddr=t.EUI64.convert("FF:FF:FF:FF:FF:FF:FF:FF"),  # global
+    Key=t.KeyData(b"ZigBeeAlliance09"),
+    TxFrameCounter=0,
+    RxFrameCounter=0,
+)
+ZSTACK_CONFIGURE_SUCCESS = b"\x55"
+
 LOGGER = logging.getLogger(__name__)
 
 
@@ -66,10 +71,18 @@ class ZNPCoordinator(zigpy.device.Device):
     def model(self):
         # There is no way to query the Z-Stack version or even the hardware at runtime.
         # Instead we have to rely on these sorts of "hints"
-        if isinstance(self.application._nib, OldNIB):
-            return "CC2531"
+        if self.application.is_cc2531:
+            model = "CC2531"
+
+            if self.application.is_zstack_home_12:
+                version = "Home 1.2"
+            else:
+                version = "3.0.1/3.0.2"
         else:
-            return "CC13X2/CC26X2"
+            model = "CC13X2/CC26X2"
+            version = "3.30.00/3.40.00/4.10.00"
+
+        return f"{model}, Z-Stack {version}"
 
 
 class ControllerApplication(zigpy.application.ControllerApplication):
@@ -85,7 +98,8 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         self._reconnect_task = asyncio.Future()
         self._reconnect_task.cancel()
 
-        self._nib = None
+        self._nib = NIB()
+        self._network_key = None
         self._concurrent_requests_semaphore = None
 
     ##################################################################
@@ -93,15 +107,26 @@ class ControllerApplication(zigpy.application.ControllerApplication):
     ##################################################################
 
     @property
-    def channel(self):
-        # This value is accessible only from the NIB struct. There does not appear to be
-        # a MT command to read it.
-
-        if self._nib is None:
-            return None
-
+    def channel(self) -> typing.Optional[int]:
         return self._nib.nwkLogicalChannel
 
+    @property
+    def channels(self) -> typing.Optional[t.Channels]:
+        return self._nib.channelList
+
+    @property
+    def pan_id(self) -> typing.Optional[t.NWK]:
+        return self._nib.nwkPanId
+
+    @property
+    def ext_pan_id(self) -> typing.Optional[t.EUI64]:
+        return self._nib.extendedPANID
+
+    @property
+    def network_key(self) -> typing.Optional[t.KeyData]:
+        # This is not a standard Zigpy property
+        return self._network_key
+
     @classmethod
     async def probe(cls, device_config: conf.ConfigType) -> bool:
         """
@@ -113,7 +138,9 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         LOGGER.debug("Probing %s", znp._port_path)
 
         try:
-            await znp.connect()
+            async with async_timeout.timeout(PROBE_TIMEOUT):
+                await znp.connect()
+
             result = True
         except Exception as e:
             result = False
@@ -135,6 +162,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         if self._znp is not None:
             self._znp.close()
+            self._znp = None
 
     async def startup(self, auto_form=False):
         """
@@ -144,12 +172,22 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         forming a network, and configuring our settings.
         """
 
+        try:
+            return await self._startup(auto_form=auto_form)
+        except Exception:
+            await self.shutdown()
+            raise
+
+    async def _startup(self, auto_form=False):
+        assert self._znp is None
+
         znp = ZNP(self.config)
-        znp.set_application(self)
         await znp.connect()
 
         # We only assign `self._znp` after it has successfully connected
         self._znp = znp
+        self._znp.set_application(self)
+
         self._bind_callbacks()
 
         # XXX: To make sure we don't switch to the wrong device upon reconnect,
@@ -157,16 +195,19 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         if self._config[conf.CONF_DEVICE][conf.CONF_DEVICE_PATH] == "auto":
             self._config[conf.CONF_DEVICE][
                 conf.CONF_DEVICE_PATH
-            ] = self._znp._uart.transport.serial.name
+            ] = self._znp._uart._transport.serial.name
 
         # Next, read out the NVRAM item that Zigbee2MQTT writes when it has configured
         # a device to make sure that our network settings will not be reset.
+        if self.is_zstack_home_12:
+            configured_nv_item = NwkNvIds.HAS_CONFIGURED_ZSTACK1
+        else:
+            configured_nv_item = NwkNvIds.HAS_CONFIGURED_ZSTACK3
+
         try:
-            is_configured = (
-                await self._znp.nvram_read(NwkNvIds.HAS_CONFIGURED_ZSTACK3)
-            ) == b"\x55"
-        except InvalidCommandResponse as e:
-            assert e.response.Status == t.Status.INVALID_PARAMETER
+            configured_value = await self._znp.nvram_read(configured_nv_item)
+            is_configured = configured_value == ZSTACK_CONFIGURE_SUCCESS
+        except KeyError:
             is_configured = False
 
         if not is_configured:
@@ -202,6 +243,45 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                 RspStatus=t.Status.SUCCESS,
             )
 
+        # Both versions of Z-Stack use this callback
+        started_as_coordinator = self._znp.wait_for_response(
+            c.ZDO.StateChangeInd.Callback(State=t.DeviceState.StartedAsCoordinator)
+        )
+
+        # The AUTOSTART startup NV item doesn't do anything.
+        if self.is_zstack_home_12:
+            # Z-Stack Home 1.2 has a simple startup sequence
+            await self._znp.request(
+                c.ZDO.StartupFromApp.Req(StartDelay=100),
+                RspState=c.zdo.StartupState.RestoredNetworkState,
+            )
+        else:
+            # Z-Stack 3 uses the BDB subsystem
+            bdb_commissioning_done = self._znp.wait_for_response(
+                c.AppConfig.BDBCommissioningNotification.Callback(
+                    partial=True, RemainingModes=c.app_config.BDBCommissioningMode.NONE
+                )
+            )
+
+            # According to the forums, this is the correct startup sequence, including
+            # the formation failure error
+            await self._znp.request_callback_rsp(
+                request=c.AppConfig.BDBStartCommissioning.Req(
+                    Mode=c.app_config.BDBCommissioningMode.NwkFormation
+                ),
+                RspStatus=t.Status.SUCCESS,
+                callback=c.AppConfig.BDBCommissioningNotification.Callback(
+                    partial=True,
+                    Status=c.app_config.BDBCommissioningStatus.NetworkRestored,
+                ),
+            )
+
+            await bdb_commissioning_done
+
+        # The startup sequence should not take forever
+        async with async_timeout.timeout(STARTUP_TIMEOUT):
+            await started_as_coordinator
+
         device_info = await self._znp.request(
             c.Util.GetDeviceInfo.Req(), RspStatus=t.Status.SUCCESS
         )
@@ -213,34 +293,13 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         # `self._register_endpoint()` adds endpoints to this device object.
         self.devices[self.ieee] = ZNPCoordinator(self, self.ieee, self.nwk)
 
-        # Start the application and wait until it's ready
-        started_as_coordinator = self._znp.wait_for_response(
-            c.ZDO.StateChangeInd.Callback(State=t.DeviceState.StartedAsCoordinator)
-        )
-
-        bdb_commissioning_done = self._znp.wait_for_response(
-            c.AppConfig.BDBCommissioningNotification.Callback(
-                partial=True, RemainingModes=c.app_config.BDBCommissioningMode.NONE
-            )
-        )
-
-        # The AUTOSTART startup NV item doesn't do anything.
-        # According to the forums, this is the correct startup sequence, including
-        # the formation failure error
-        await self._znp.request_callback_rsp(
-            request=c.AppConfig.BDBStartCommissioning.Req(
-                Mode=c.app_config.BDBCommissioningMode.NwkFormation
-            ),
+        # Give our Zigpy device a valid node descriptor
+        node_descriptor_rsp = await self._znp.request_callback_rsp(
+            request=c.ZDO.NodeDescReq.Req(DstAddr=0x0000, NWKAddrOfInterest=0x0000),
             RspStatus=t.Status.SUCCESS,
-            callback=c.AppConfig.BDBCommissioningNotification.Callback(
-                partial=True, Status=c.app_config.BDBCommissioningStatus.NetworkRestored
-            ),
+            callback=c.ZDO.NodeDescRsp.Callback(Src=0x0000, NWK=0x0000, partial=True),
         )
-
-        # The startup sequence should not take forever
-        async with async_timeout.timeout(20):
-            # These often arrive in random order
-            await asyncio.gather(started_as_coordinator, bdb_commissioning_done)
+        self.zigpy_device.node_desc = node_descriptor_rsp.NodeDescriptor
 
         # Get the currently active endpoints
         endpoints = await self._znp.request_callback_rsp(
@@ -277,85 +336,45 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         # Now that we know what device we are, set the max concurrent requests
         if self.znp_config[conf.CONF_MAX_CONCURRENT_REQUESTS] == "auto":
-            max_concurrent_requests = 2 if self.is_cc2531 else 16
+            max_concurrent_requests = 2 if isinstance(self._nib, CC2531NIB) else 16
         else:
             max_concurrent_requests = self.znp_config[conf.CONF_MAX_CONCURRENT_REQUESTS]
 
         self._concurrent_requests_semaphore = asyncio.Semaphore(max_concurrent_requests)
 
         LOGGER.info(
-            "Using channel mask %s, currently on channel %d."
-            " Limiting concurrent requests to %d",
-            self.channels,
+            "Currently on channel %d. Limiting concurrent requests to %d",
             self.channel,
             max_concurrent_requests,
         )
 
-    async def update_network(
-        self,
-        *,
-        channel: typing.Optional[t.uint8_t] = None,
-        channels: typing.Optional[t.Channels] = None,
-        extended_pan_id: typing.Optional[t.ExtendedPanId] = None,
-        network_key: typing.Optional[t.KeyData] = None,
-        pan_id: typing.Optional[t.PanId] = None,
-        tc_address: typing.Optional[t.EUI64] = None,
-        tc_link_key: typing.Optional[t.KeyData] = None,
-        update_id: int = 0,
-        reset: bool = True,
-    ):
+    async def update_network_channel(self, channel: t.uint8_t):
         """
-        Updates network settings at runtime, after the application has started.
+        Changes the network channel, increments the beacon update ID, and waits until
+        the changes take effect.
+
+        Does nothing if the new channel is the same as the old.
         """
 
-        if (
-            channel is not None
-            and channels is not None
-            and not t.Channels.from_channel_list([channel]) & channels
-        ):
-            raise ValueError("Channel does not overlap with channel mask")
+        if self.channel == channel:
+            return
 
-        if tc_link_key is not None:
-            LOGGER.warning("Trust center link key in config is not yet supported")
+        await self._znp.request(
+            request=c.ZDO.MgmtNWKUpdateReq.Req(
+                Dst=0x0000,
+                DstAddrMode=t.AddrMode.NWK,
+                Channels=t.Channels.from_channel_list([channel]),
+                ScanDuration=0xFE,  # switch channels
+                ScanCount=0,
+                NwkManagerAddr=0x0000,
+            ),
+            RspStatus=t.Status.SUCCESS,
+        )
 
-        if tc_address is not None:
-            LOGGER.warning("Trust center address in config is not yet supported")
-
-        if channels is not None:
-            await self._znp.nvram_write(NwkNvIds.CHANLIST, channels)
-            await self._znp.request(
-                c.AppConfig.BDBSetChannel.Req(IsPrimary=True, Channel=channels),
-                RspStatus=t.Status.SUCCESS,
-            )
-            await self._znp.request(
-                c.AppConfig.BDBSetChannel.Req(
-                    IsPrimary=False, Channel=t.Channels.NO_CHANNELS
-                ),
-                RspStatus=t.Status.SUCCESS,
-            )
-
-        if channel is not None:
-            # XXX: We modify the logical channel value directly in the NIB.
-            #      Is there no better way?
-            nib = parse_nib(await self._znp.nvram_read(NwkNvIds.NIB))
-            nib.nwkLogicalChannel = channel
-            await self._znp.nvram_write(NwkNvIds.NIB, nib.serialize())
-
-        if pan_id is not None:
-            await self._znp.nvram_write(NwkNvIds.PANID, pan_id)
-
-        if extended_pan_id is not None:
-            # There is no Util request to do this
-            await self._znp.nvram_write(NwkNvIds.EXTENDED_PAN_ID, extended_pan_id)
-
-        if network_key is not None:
-            await self._znp.nvram_write(NwkNvIds.PRECFGKEY, network_key)
-            await self._znp.nvram_write(NwkNvIds.PRECFGKEYS_ENABLE, t.Bool(True))
-
-        if reset:
-            # We have to reset afterwards
-            await self._reset()
+        # The above command takes a few seconds to work
+        while self.channel != channel:
             await self._load_device_info()
+            await asyncio.sleep(1)
 
     async def form_network(self):
         """
@@ -363,10 +382,33 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         PAN ID, and extended PAN ID.
         """
 
-        # Delete any existing HAS_CONFIGURED_ZSTACK3 NV item. This may fail.
-        await self._znp.request(
-            c.SYS.OSALNVDelete.Req(Id=NwkNvIds.HAS_CONFIGURED_ZSTACK3, ItemLen=1)
-        )
+        # First, make the settings consistent and randomly generate missing values
+        channel = self.config[conf.CONF_NWK][conf.CONF_NWK_CHANNEL]
+        channels = self.config[conf.CONF_NWK][conf.CONF_NWK_CHANNELS]
+        pan_id = self.config[conf.CONF_NWK][conf.CONF_NWK_PAN_ID]
+        extended_pan_id = self.config[conf.CONF_NWK][conf.CONF_NWK_EXTENDED_PAN_ID]
+        network_key = self.config[conf.CONF_NWK][conf.CONF_NWK_KEY]
+
+        if pan_id is None:
+            # Let Z-Stack pick one at random, hopefully not conflicting with others
+            pan_id = t.uint16_t(0xFFFF)
+
+        if extended_pan_id is None:
+            # It's not documented whether or not Z-Stack will pick this randomly as well
+            # if a value of 00:00:00:00:00:00:00:00 is provided but the chances of a
+            # collision using `os.urandom` are astronomically small
+            extended_pan_id = ExtendedPanId(os.urandom(8))
+
+        if network_key is None:
+            network_key = t.KeyData(os.urandom(16))
+
+        # Override `channels` with a single channel if one is explicitly set
+        if channel is not None:
+            channels = t.Channels.from_channel_list([channel])
+
+        # Delete any existing HAS_CONFIGURED_ZSTACK* NV items. One (or both) may fail.
+        await self._znp.nvram_delete(NwkNvIds.HAS_CONFIGURED_ZSTACK1)
+        await self._znp.nvram_delete(NwkNvIds.HAS_CONFIGURED_ZSTACK3)
 
         # Instruct Z-Stack to reset everything on the next boot
         await self._znp.nvram_write(
@@ -378,106 +420,115 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         await self._reset()
 
         # Now that we've cleared everything, write back our Z-Stack settings
-        await self._write_stack_settings(reset_if_changed=False)
-
-        pan_id = self.config[conf.CONF_NWK][conf.CONF_NWK_PAN_ID]
-
-        if pan_id is None:
-            # Let Z-Stack pick one at random, hopefully not conflicting with others
-            pan_id = t.uint16_t(0xFFFF)
-
-        extended_pan_id = self.config[conf.CONF_NWK][conf.CONF_NWK_EXTENDED_PAN_ID]
-
-        if extended_pan_id is None:
-            # It's not documented whether or not Z-Stack will pick this randomly as well
-            # if a value of 00:00:00:00:00:00:00:00 is provided but the chances of a
-            # collision using `os.urandom` are astronomically small
-            extended_pan_id = ExtendedPanId(os.urandom(8))
-
         LOGGER.debug("Updating network settings")
 
-        # Update the network settings.
-        # Not resetting before we form the network is important!
-        await self.update_network(
-            # We don't set the channel in here because it will have no effect
-            channels=self.config[conf.CONF_NWK][conf.CONF_NWK_CHANNELS],
-            pan_id=pan_id,
-            extended_pan_id=extended_pan_id,
-            network_key=t.KeyData(os.urandom(16)),
-            reset=False,
+        await self._write_stack_settings(reset_if_changed=False)
+        await self._znp.nvram_write(NwkNvIds.PANID, pan_id, create=True)
+        await self._znp.nvram_write(
+            NwkNvIds.APS_USE_EXT_PANID, extended_pan_id, create=True
         )
+        await self._znp.nvram_write(NwkNvIds.PRECFGKEY, network_key, create=True)
+        await self._znp.nvram_write(
+            NwkNvIds.PRECFGKEYS_ENABLE, t.Bool(True), create=True
+        )
+        await self._znp.nvram_write(NwkNvIds.CHANLIST, channels, create=True)
+
+        # Z-Stack Home 1.2 doesn't have the BDB subsystem
+        if not self.is_zstack_home_12:
+            await self._znp.request(
+                c.AppConfig.BDBSetChannel.Req(IsPrimary=True, Channel=channels),
+                RspStatus=t.Status.SUCCESS,
+            )
+            await self._znp.request(
+                c.AppConfig.BDBSetChannel.Req(
+                    IsPrimary=False, Channel=t.Channels.NO_CHANNELS
+                ),
+                RspStatus=t.Status.SUCCESS,
+            )
 
         # Finally, form the network
         LOGGER.debug("Forming the network")
 
-        # Form the network and capture expected progress messages so they don't appear
-        # as warnings within logs
+        started_as_coordinator = self._znp.wait_for_response(
+            c.ZDO.StateChangeInd.Callback(State=t.DeviceState.StartedAsCoordinator)
+        )
+
+        # Handle the startup progress messages
         async with self._znp.capture_responses(
             [
                 c.ZDO.StateChangeInd.Callback(
                     State=t.DeviceState.StartingAsCoordinator
                 ),
-                c.ZDO.StateChangeInd.Callback(State=t.DeviceState.StartedAsCoordinator),
                 c.AppConfig.BDBCommissioningNotification.Callback(
-                    partial=True, Status=c.app_config.BDBCommissioningStatus.InProgress
+                    partial=True,
+                    Status=c.app_config.BDBCommissioningStatus.InProgress,
                 ),
             ]
         ):
-            bdb_commissioning_rsp = await self._znp.request_callback_rsp(
-                request=c.AppConfig.BDBStartCommissioning.Req(
-                    Mode=c.app_config.BDBCommissioningMode.NwkFormation
-                ),
-                RspStatus=t.Status.SUCCESS,
-                callback=c.AppConfig.BDBCommissioningNotification.Callback(
-                    partial=True, RemainingModes=c.app_config.BDBCommissioningMode.NONE
-                ),
-            )
+            if not self.is_zstack_home_12:
+                # Z-Stack 3 uses the BDB subsystem
+                commissioning_rsp = await self._znp.request_callback_rsp(
+                    request=c.AppConfig.BDBStartCommissioning.Req(
+                        Mode=c.app_config.BDBCommissioningMode.NwkFormation
+                    ),
+                    RspStatus=t.Status.SUCCESS,
+                    callback=c.AppConfig.BDBCommissioningNotification.Callback(
+                        partial=True,
+                        RemainingModes=c.app_config.BDBCommissioningMode.NONE,
+                    ),
+                    timeout=NETWORK_COMMISSIONING_TIMEOUT,
+                )
 
-        if bdb_commissioning_rsp.Status != c.app_config.BDBCommissioningStatus.Success:
-            raise RuntimeError(f"Network formation failed: {bdb_commissioning_rsp}")
+                if (
+                    commissioning_rsp.Status
+                    != c.app_config.BDBCommissioningStatus.Success
+                ):
+                    raise RuntimeError(f"Network formation failed: {commissioning_rsp}")
+            else:
+                await self._znp.nvram_write(
+                    NwkNvIds.TCLK_SEED,
+                    value=DEFAULT_TC_LINK_KEY,
+                    create=True,
+                )
 
-        LOGGER.debug("Waiting for the network NIB to be populated")
+                # In Z-Stack 1.2.2, StartupFromApp actually does what it says
+                await self._znp.request(
+                    c.ZDO.StartupFromApp.Req(StartDelay=100),
+                    RspState=c.zdo.StartupState.NewNetworkState,
+                )
 
-        # Even though the device is "ready" at this point, for some reason it takes
-        # a few more seconds for the NIB to update with our correct logical channel
+            # Both versions still end with this callback
+            await started_as_coordinator
+
+        LOGGER.debug("Waiting for the NIB to be populated")
+
+        # Even though the device says it is "ready" at this point, it takes a few more
+        # seconds for `_NIB.nwkState` to switch from `NwkState.NWK_INIT`. There does
+        # not appear to be any user-facing MT command to read this information.
         while True:
-            nib = parse_nib(await self._znp.nvram_read(NwkNvIds.NIB))
+            try:
+                nib = parse_nib(await self._znp.nvram_read(NwkNvIds.NIB))
 
-            # Usually this works after the first attempt
-            if nib.nwkLogicalChannel:
-                break
+                LOGGER.debug("Current NIB is %s", nib)
+
+                # Usually this works after the first attempt
+                if nib.nwkLogicalChannel != 0 and nib.nwkPanId != 0xFFFE:
+                    break
+            except KeyError:
+                pass
 
             await asyncio.sleep(1)
 
-        # Only at this point can we update our logical channel
-        channel = self.config[conf.CONF_NWK][conf.CONF_NWK_CHANNEL]
+        # Create the NV item that keeps track of whether or not we're fully configured.
+        if self.is_zstack_home_12:
+            configured_nv_item = NwkNvIds.HAS_CONFIGURED_ZSTACK1
+        else:
+            configured_nv_item = NwkNvIds.HAS_CONFIGURED_ZSTACK3
 
-        if channel is not None and channel != nib.nwkLogicalChannel:
-            LOGGER.debug(
-                "Z-Stack started with channel %d. Updating to %d.",
-                nib.nwkLogicalChannel,
-                channel,
-            )
-            await self.update_network(channel=channel)
-
-        # Create the NV item that keeps track of whether or not we're configured.
-        # This is the NV item used by Zigbee2MQTT, pulled in from zigbee-shepherd, and
-        # allows for this device to be used with Zigbee2MQTT as well.
-        osal_create_rsp = await self._znp.request(
-            c.SYS.OSALNVItemInit.Req(
-                Id=NwkNvIds.HAS_CONFIGURED_ZSTACK3, ItemLen=1, Value=b"\x55"
-            )
+        await self._znp.nvram_write(
+            configured_nv_item, ZSTACK_CONFIGURE_SUCCESS, create=True
         )
 
-        if osal_create_rsp.Status not in (t.Status.SUCCESS, t.Status.NV_ITEM_UNINIT):
-            raise RuntimeError(
-                f"Network formation failed: could not create"
-                f" HAS_CONFIGURED_ZSTACK3 NV item: {osal_create_rsp}"
-            )  # pragma: no cover
-
-        # Initializing the item doesn't guarantee that it holds this exact value
-        await self._znp.nvram_write(NwkNvIds.HAS_CONFIGURED_ZSTACK3, b"\x55")
-
         # Finally, reset once more to reset the device back to a "normal" state so that
         # we can continue the normal application startup sequence.
         await self._reset()
@@ -631,9 +682,50 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         if response.Status != t.Status.SUCCESS:
             raise RuntimeError(f"Permit join response failure: {response}")
 
+    async def permit_with_key(self, node: t.EUI64, code: bytes, time_s=60):
+        """
+        Permits a new device to join with the given IEEE and Install Code.
+        """
+
+        key = zigpy.util.convert_install_code(code)
+        install_code_format = c.app_config.InstallCodeFormat.KeyDerivedFromInstallCode
+
+        if key is None:
+            raise ValueError(f"Invalid install code: {code!r}")
+
+        await self._znp.request(
+            c.AppConfig.BDBAddInstallCode.Req(
+                InstallCodeFormat=install_code_format,
+                IEEE=node,
+                InstallCode=t.Bytes(key),
+            ),
+            RspStatus=t.Status.SUCCESS,
+        )
+
+        # Temporarily only allow joins that use an install code
+        await self._znp.request(
+            c.AppConfig.BDBSetJoinUsesInstallCodeKey.Req(
+                BdbJoinUsesInstallCodeKey=True
+            ),
+            RspStatus=t.Status.SUCCESS,
+        )
+
+        try:
+            await self.permit(time_s)
+            await asyncio.sleep(time_s)
+        finally:
+            # Revert back to normal. The BDB config is not persistent so if this request
+            # fails, we will be back to normal the next time Z-Stack resets.
+            await self._znp.request(
+                c.AppConfig.BDBSetJoinUsesInstallCodeKey.Req(
+                    BdbJoinUsesInstallCodeKey=False
+                ),
+                RspStatus=t.Status.SUCCESS,
+            )
+
     def connection_lost(self, exc):
         """
-        Propagated up from lower layers (UART and ZNP) when the connection is lost.
+        Propagated up from UART through ZNP when the connection is lost.
         Spawns the auto-reconnect task.
         """
 
@@ -645,10 +737,9 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             return
 
         # Reconnect in the background using our previously-detected port.
-        LOGGER.debug("Starting background reconnection task")
-
-        self._reconnect_task.cancel()
-        self._reconnect_task = asyncio.create_task(self._reconnect())
+        if self._reconnect_task.done():
+            LOGGER.debug("Restarting background reconnection task")
+            self._reconnect_task = asyncio.create_task(self._reconnect())
 
     #####################################################
     # Z-Stack message callbacks attached during startup #
@@ -668,11 +759,13 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         # ZDO requests need to be handled explicitly, one by one
         self._znp.callback_for_response(
-            c.ZDO.EndDeviceAnnceInd.Callback(partial=True), self.on_zdo_device_announce,
+            c.ZDO.EndDeviceAnnceInd.Callback(partial=True),
+            self.on_zdo_device_announce,
         )
 
         self._znp.callback_for_response(
-            c.ZDO.TCDevInd.Callback.Callback(partial=True), self.on_zdo_tc_device_join,
+            c.ZDO.TCDevInd.Callback.Callback(partial=True),
+            self.on_zdo_tc_device_join,
         )
 
         self._znp.callback_for_response(
@@ -765,12 +858,12 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         return self.devices[self.ieee]
 
     @property
-    def is_cc2531(self) -> bool:
-        """
-        There really are only two ZNP radios: the cheap CC2531 and the high-power ones.
-        """
+    def is_zstack_home_12(self) -> bool:
+        return self._znp._version.MinorRel == 6
 
-        return isinstance(self._nib, OldNIB)
+    @property
+    def is_cc2531(self) -> bool:
+        return isinstance(self._nib, CC2531NIB)
 
     @property
     def znp_config(self) -> conf.ConfigType:
@@ -795,6 +888,8 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             NwkNvIds.CONCENTRATOR_RC: t.Bool(True),
             NwkNvIds.SRC_RTG_EXPIRY_TIME: t.uint8_t(255),
             NwkNvIds.NWK_CHILD_AGE_ENABLE: t.Bool(False),
+            # Default is 20 in Z-Stack 3.0.1, 30 in Z-Stack 3/4
+            NwkNvIds.BCAST_DELIVERY_TIME: t.uint8_t(30),
             # We want to receive all ZDO callbacks to proxy them back to zigpy
             NwkNvIds.ZDO_DIRECT_CB: t.Bool(True),
         }
@@ -858,22 +953,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         assert set(zdo_kwargs) == set(field_names)
 
         # Type cast all of the field args and kwargs
-        zdo_args = []
-
-        for name, field_type in zip(field_names, field_types):
-            zdo_arg = zdo_kwargs[name]
-
-            if issubclass(field_type, ZigpyStruct) and hasattr(ZigpyStruct, "_fields"):
-                # Old-style zigpy structs do not have "copy constructors"
-                new_obj = field_type()
-
-                for field_name, _ in new_obj._fields:
-                    setattr(new_obj, field_name, getattr(zdo_arg, field_name))
-            else:
-                new_obj = field_type(zdo_arg)
-
-            zdo_args.append(zdo_arg)
-
+        zdo_args = [t(zdo_kwargs[name]) for name, t in zip(field_names, field_types)]
         message = t.serialize_list([t.uint8_t(tsn)] + zdo_args)
 
         LOGGER.debug("Pretending we received a ZDO message: %s", message)
@@ -903,8 +983,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             )
 
             try:
-                await self.startup()
-
+                await self._startup()
                 return
             except Exception as e:
                 LOGGER.error("Failed to reconnect", exc_info=e)
@@ -960,18 +1039,11 @@ class ControllerApplication(zigpy.application.ControllerApplication):
     async def _load_device_info(self):
         # Parsing the NIB struct gives us access to low-level info, like the channel
         self._nib = parse_nib(await self._znp.nvram_read(NwkNvIds.NIB))
-        LOGGER.debug("Parsed NIB: %s", self._nib)
+        self._network_key, _ = t.KeyData.deserialize(
+            await self._znp.nvram_read(NwkNvIds.PRECFGKEY)
+        )
 
-        # Util.GetNvInfo reads all of these but it has an endianness bug with CHANLIST
-        self._pan_id, _ = t.PanId.deserialize(
-            await self._znp.nvram_read(NwkNvIds.PANID)
-        )
-        self._channels, _ = t.Channels.deserialize(
-            await self._znp.nvram_read(NwkNvIds.CHANLIST)
-        )
-        self._ext_pan_id, _ = t.EUI64.deserialize(
-            await self._znp.nvram_read(NwkNvIds.EXTENDED_PAN_ID)
-        )
+        LOGGER.debug("Parsed NIB: %s", self._nib)
 
     async def _reset(self):
         """
